/*
Title- Exploring Correlation for Indirect Branch Prediction
Authors- Nikunj Bhansali - NC State University- nsbhansa@ncsu.edu
	 Chintan Panirwala - NC State University-cdpanirw@ncsu.edu
         Huiyang Zhou- NC State University- hzhou@ncsu.edu

Date-  May 2011

Code has been inspired by LTAGE predictor by Dr. Andre Seznec.
*/
#include <inttypes.h>
#include <math.h>

class branch_update {
  bool _direction_prediction;
  unsigned int _target_prediction;
  
 public:
  bool direction_prediction () { return _direction_prediction; }
  void direction_prediction (bool b) { _direction_prediction = b; }
  
  uint32_t target_prediction () { return _target_prediction; }
  void target_prediction (unsigned int t) { _target_prediction = t; }
  
  branch_update (void) :
    _direction_prediction(false), _target_prediction(0) {}
};

class branch_predictor {
 public:
  virtual branch_update *predict_fetch (unsigned int pc)=0;
  virtual void update (branch_update *, bool, unsigned int, int) {}
  virtual ~branch_predictor (void) {}
};

//Total storage for the submitted predictor

//ITTAGE (gtable[] and the number of entries in each table is defined in logg[])
//Table T0: 0 bits (no memory is allocated for gtable[0])
//Table T1: 2048 entries, each entry has a 7-bit tag, 2-bit u field, 32-bit target address. Total 2048x(7+2+32) = 83968 bits
//Table T2: 1024 entries, each entry has a 7-bit tag, 2-bit u field, 1-bit alt field, 32-bit target address. Total: 43008 bits
//Table T3: 4096 entries, each entry has a 8-bit tag, 2-bit u field, 1-bit alt field, 32-bit target address. Total:176128 bits
//Table T4: 2048 entries, each entry has a 8-bit tag, 2-bit u field, 1-bit alt field, 32-bit target address. Total:88064 bits
//Table T5: 512 entries, each entry has a 9-bit tag, 2-bit u field, 1-bit alt field, 32-bit target address. Total:22528 bits
//Table T6: 512 entries, each entry has a 10-bit tag, 2-bit u field, 1-bit alt field, 32-bit target address. Total:23040 bits
//Table T7: 512 entries, each entry has an 11-bit tag, 2-bit u field, 1-bit alt field, 32-bit target address. Total:23552 bits
//Table T8: 512 entries, each entry has a 12-bit tag, 2-bit u field, 1-bit alt field, 32-bit target address. Total:24064 bits
//Table T9: 512 entries, each entry has a 12-bit tag, 2-bit u field, 1-bit alt field, 32-bit target address. Total:24064 bits
//Table T10: 64 entries, each entry has a 13-bit tag, 2-bit u field, 1-bit alt field, 32-bit target address. Total:3072 bits
//Table T11: 64 entries, each entry has a 14-bit tag, 2-bit u field, 1-bit alt field, 32-bit target address. Total:3136 bits
//Table T12: 64 entries, each entry has a 15-bit tag, 2-bit u field, 1-bit alt field, 32-bit target address. Total:3200 bits

//History bits: 2x640 = 1280. 

//Misc: 2*16 path history bits + 4 bits for  USE_ALT_ON_NA + 19 bits for the TICK counter = 55 bits
//We used the same history length and same history management as in LTAGE conditional branch predictor. Therefore, we use the same storage cost.

//Hard-to-predict branch table (cache_table)
//32 entries, each entry has a 32-bit tag, 2-bit u field, 4-bit hlen field Total:1216 bits 

//Address-target correlation table (exec_table)
//26 entries, each entry has a 32-bit tag, a 5-bit LRU field, 10*10 bits for storing addresses and 32*10 bits for targets. Total = 11882 bits 

//Overall cost = 532202 + 55 = 532257 bits

//Compared to the budget of 65kB = 65*1024*8 = 532480, the difference (223 bits) is the range to cover the missing storage cost (if there is any).
//Furthermore, if we use the following definition of NUM_PAIR and EXE_T_SIZE for exec_table (i.e., 25 enries with each entry containing 9 pairs of addresses and targets), the overall storage cost is 530750 bits, providing an error range of 1730 bits. The difference in the overall MPPKI is only 0.0365%. 

//#define NUM_PAIR 9             // Number of addr-target pairs in each entyr of the addr-target correlation table
//#define EXEC_T_SIZE 25          // Number of entries in the addr-target correlation table

#define NUM_PAIR 10             // Number of addr-target pairs in each entyr of the addr-target correlation table
#define EXEC_T_SIZE 26          // Number of entries in the addr-target correlation table

#define NHIST 12		//12 tagged components
#define LOGB 14			// log of the number of entries in the base bimodal predictor
#define HYSTSHIFT 2		// sharing an hysteris bit between 4 bimodal predictor entries
#define LOGG (LOGB-3)		// base 2 logarithm of number of entries  on each tagged component
#define TBITS 7			// minimum tag width (shortest history length table)
#define MINHIST 4		// shortest history length, same as in LTAGE
#define MAXHIST 640		// longest history length, same as in LTAGE

typedef uint32_t address_t;

#define BUFFERHIST 128 * 1024   //same as in LTAGE
//size of the history buffer: to allow fast management we use a very large buffer //we replace a global shift by a pointer management

// utility class for index computation

// this is the cyclic shift register for folding 
// a long global history into a smaller number of bits; see P. Michaud's PPM-like predictor at CBP-1
class folded_history
{
public:
  unsigned comp;
  int CLENGTH;
  int OLENGTH;
  int OUTPOINT;

  folded_history ()
  {
  }

  void init (int original_length, int compressed_length)
  {//It is the same code as in LTAGE predictor, thereby the same way of counting the storage cost
    comp = 0;
    OLENGTH = original_length;
    CLENGTH = compressed_length;
    OUTPOINT = OLENGTH % CLENGTH;
  }

  void update (uint8_t * h)
  {
    comp = (comp << 1) | h[0];
    comp ^= h[OLENGTH] << OUTPOINT;
    comp ^= (comp >> CLENGTH);
    comp &= (1 << CLENGTH) - 1;
  }
};


class my_predictor:public branch_predictor
{
 public:
  branch_update u;//It is the same code as in LTAGE predictor, thereby the same way of counting the storage cost

  cbp3_uop_dynamic_t bi;

  class cache_entry
  {
  public :
    unsigned int pc;
    int mspcount,hist_len;
    cache_entry()
      {
	pc=0;
	mspcount=0;
	hist_len=1;
      }	
  };	

  class gentry			// TAGE global table entry
  {
  public:
    uint16_t tag;
    int8_t u;
    uint32_t targetaddr;
    unsigned int alt;
    gentry ()
    {
      targetaddr=0;
      tag = 0;
      u = 0;
      alt = false;
    }
    void init()
    {
      targetaddr=0;
      tag = 0;
      u = 0;
      alt = false;
    }
  };
  class cache_execute_predict   //an entry in the addr-target correlation table
  {
   public :
    unsigned int pc;
    unsigned int lru; //LRU replacement
    int sr[NUM_PAIR];//10 bits each for storing source value
    unsigned int pr[NUM_PAIR];//32 bits each
    cache_execute_predict()
    {
      pc=0;
      lru = 0;
      for (int i=0;i<NUM_PAIR;i++)
      {
	sr[i]=0;
	pr[i]=0;
      } 
    }
  };


  int USE_ALT_ON_NA;		// "Use alternate prediction on newly allocated":  a 4-bit counter  to determine whether the newly allocated entries should be considered as  valid or not for delivering  the prediction

  int TICK, LOGTICK;		//control counter for the smooth resetting of useful counters

  int phist_pred;			// use a path history as on  the OGEHL predictor
//for managing global history  
  uint8_t *GHIST_pred;
  uint8_t *ghist_pred;
  int ptghist_pred;

  int phistos_pred;			// path history including kernel activity
// for managing global history including kernel activity
  uint8_t *GHISTOS_pred;
  uint8_t *ghistos_pred;
  int ptghistos_pred;

  int phist;			// use a path history as on  the OGEHL predictor
//for managing global history  
  uint8_t *GHIST;
  uint8_t *ghist;
  int ptghist;

  int phistos;			// path history including kernel activity
// for managing global history including kernel activity
  uint8_t *GHISTOS;
  uint8_t *ghistos;
  int ptghistos;


  //The following folded history vaiables are based on history bits for efficient index & tag computation. No real storage is needed. 
  //It is the same code as in LTAGE predictor, thereby the same way of counting the storage cost

  folded_history ch_i[NHIST + 1];	//utility for computing TAGE indices 
  folded_history ch_t[2][NHIST + 1];	//utility for computing TAGE tags

  folded_history ch_ios[NHIST + 1];	//utility for computing TAGE indices on kernel branches
  folded_history ch_tos[2][NHIST + 1];	//utility for computing TAGE tags on kernel branches

  folded_history ch_i_pred[NHIST + 1];	//utility for computing TAGE indices
  folded_history ch_t_pred[2][NHIST + 1];	//utility for computing TAGE tags

  folded_history ch_ios_pred[NHIST + 1];	//utility for computing TAGE indices on kernel branches
  folded_history ch_tos_pred[2][NHIST + 1];	//utility for computing TAGE tags on kernel branches


  //Actual storage for TAGE
  gentry *gtable[NHIST + 1];	// tagged TAGE tables

  //Actual storage for hard-to-predict branches
  cache_entry *cache_table;

  //actual storage for address-target correlation table
  cache_execute_predict *exec_table;

  //The following array variables are fixed values based on number of table entries, number of tag bits in each table, history lengths used for each table. So, no actual storage
  //It is the same code as in LTAGE predictor, thereby the same way of counting the storage cost
  int TB[NHIST + 1];		// tag width for the different tagged tables
  int m[NHIST + 1];		// used for storing the history lengths
  int logg[NHIST + 1];		// log of number entries of the different tagged tables

  //The following variables are the indices and tags based on pc and history.They are computed, rather than stored.
  //It is the same code as in LTAGE predictor, thereby the same way of counting the storage cost
  int GI[NHIST + 1];		// indexes to the different tables are computed only once  
  int GTAG[NHIST + 1];		// tags for the different tables are computed only once  
  int BI;			// index of the bimodal table
  int Seed;			// for the pseudo-random number generator

  uint32_t alttarget;           // alternate target prediction

  int HitBank;			// longest matching bank
  int AltBank;			// alternate matching bank
  uint32_t target_pred;		// prection for indirect target

  my_predictor (void)
  {
    USE_ALT_ON_NA = 0;
    Seed = 0;
    LOGTICK = 19;		//log of the period for useful tag smooth resetting
    TICK = (1 << (LOGTICK - 1));	//initialize the resetting counter to the half of the period
    phist = 0;
    phistos = 0;
    GHIST = (uint8_t *) malloc(BUFFERHIST);
    GHISTOS = (uint8_t *) malloc(BUFFERHIST);
    ghist = GHIST;
    ghistos = GHISTOS;

    for (int i = 0; i < BUFFERHIST; i++)
      ghist[i] = 0;

    for (int i = 0; i < BUFFERHIST; i++)
      ghistos[i] = 0;

    ptghist = 0;
    ptghistos = 0;

    phist_pred = 0;
    phistos_pred = 0;
    GHIST_pred = (uint8_t *) malloc(BUFFERHIST);
    GHISTOS_pred = (uint8_t *) malloc(BUFFERHIST);
    ghist_pred = GHIST_pred;
    ghistos_pred = GHISTOS_pred;

    for (int i = 0; i < BUFFERHIST; i++)
      ghist_pred[i] = 0;

    for (int i = 0; i < BUFFERHIST; i++)
      ghistos_pred[i] = 0;

    ptghist_pred = 0;
    ptghistos_pred = 0;

    // computes the geometric history lengths   
    m[1] = MINHIST;
    m[NHIST] = MAXHIST;
    for (int i = 2; i <= NHIST; i++)
      {
	m[i] =
	  (int) (((double) MINHIST *
		  pow ((double) (MAXHIST) / (double) MINHIST,
		       (double) (i - 1) / (double) ((NHIST - 1)))) + 0.5);
      }


//widths of the partial tags
    TB[1] = TBITS;
    TB[2] = TBITS;
    TB[3] = TBITS + 1;
    TB[4] = TBITS + 1;
    TB[5] = TBITS + 2;
    TB[6] = TBITS + 3;
    TB[7] = TBITS + 4;
    TB[8] = TBITS + 5;
    TB[9] = TBITS + 5;
    TB[10] = TBITS + 6;
    TB[11] = TBITS + 7;
    TB[12] = TBITS + 8;

// log2 of number entries in the tagged components
    for (int i = 1; i <= 1; i++)
      logg[i] = LOGG - 0;
     for (int i = 2; i <= 2; i++)
       logg[i] = LOGG - 1;
    for (int i = 3; i <= 3; i++)
      logg[i] = LOGG + 1;
    for (int i = 4; i <= 4; i++)
      logg[i] = LOGG ; //;
    for (int i = 5; i <= 9; i++)
      logg[i] = LOGG - 2;
    for (int i = 10; i <= 10; i++)
      logg[i] = LOGG - 5;
    for (int i = 11; i <= 12; i++)
      logg[i] = LOGG - 5;
//initialisation of index and tag computation functions
    for (int i = 1; i <= NHIST; i++)
      {
	ch_i[i].init (m[i], (logg[i]));
	ch_t[0][i].init (ch_i[i].OLENGTH, TB[i]);
	ch_t[1][i].init (ch_i[i].OLENGTH, TB[i] - 1);
      }
    for (int i = 1; i <= NHIST; i++)
      {
	ch_ios[i].init (m[i], (logg[i]));
	ch_tos[0][i].init (ch_i[i].OLENGTH, TB[i]);
	ch_tos[1][i].init (ch_i[i].OLENGTH, TB[i] - 1);
      }
    for (int i = 1; i <= NHIST; i++)
      {
	ch_i_pred[i].init (m[i], (logg[i]));
	ch_t_pred[0][i].init (ch_i[i].OLENGTH, TB[i]);
	ch_t_pred[1][i].init (ch_i[i].OLENGTH, TB[i] - 1);
      }
    for (int i = 1; i <= NHIST; i++)
      {
	ch_ios_pred[i].init (m[i], (logg[i]));
	ch_tos_pred[0][i].init (ch_i[i].OLENGTH, TB[i]);
	ch_tos_pred[1][i].init (ch_i[i].OLENGTH, TB[i] - 1);
      }

    cache_table= new cache_entry[32];

    exec_table = new cache_execute_predict[EXEC_T_SIZE];//28]; //[16];//8];    

    for(int i = 0; i < EXEC_T_SIZE; i++)//16; i++)//8; i++)
      exec_table[i].lru = i;

    for (int i = 1; i <= NHIST; i++)
      {
	gtable[i] = new gentry[1 << (logg[i])];
      }

        #define PRINTCHAR
#ifdef PRINTCHAR
//for printing predictor characteristics
    int NBENTRY = 0;

    int STORAGESIZE = 0;
    for (int i = 1; i <= NHIST; i++)
      {

	fprintf (stderr, "%d ", TB[i]);

	STORAGESIZE += (1 << logg[i]) * (2 + 33 + TB[i]);
	NBENTRY += (1 << logg[i]);


      } fprintf (stderr, "\n");

  
    for (int i = 1; i <= NHIST; i++)
      {
	fprintf (stderr, "%d ", m[i]);

      }
    fprintf (stderr, "\n");

    STORAGESIZE+=1216;
    fprintf (stderr,
	     "NHIST= %d; MINHIST= %d; MAXHIST= %d; STORAGESIZE= %d; NBENTRY= %d\n",
	     NHIST, MINHIST, MAXHIST, STORAGESIZE, NBENTRY);
#endif
  }

void init()
{
    USE_ALT_ON_NA = 0;
    Seed = 0;
    LOGTICK = 19;		//log of the period for useful tag smooth resetting
    TICK = (1 << (LOGTICK - 1));	//initialize the resetting counter to the half of the period
    phist = 0;
    phistos = 0;

    ghist = GHIST;
    ghistos = GHISTOS;

    for (int i = 0; i < BUFFERHIST; i++)
      ghist[i] = 0;

    for (int i = 0; i < BUFFERHIST; i++)
      ghistos[i] = 0;

    ptghist = 0;
    ptghistos = 0;

    phist_pred = 0;
    phistos_pred = 0;
 
    ghist_pred = GHIST_pred;
    ghistos_pred = GHISTOS_pred;

    for (int i = 0; i < BUFFERHIST; i++)
      ghist_pred[i] = 0;

    for (int i = 0; i < BUFFERHIST; i++)
      ghistos_pred[i] = 0;

    ptghist_pred = 0;
    ptghistos_pred = 0;

    // computes the geometric history lengths   
    m[1] = MINHIST;
    m[NHIST] = MAXHIST;
    for (int i = 2; i <= NHIST; i++)
      {
	m[i] =
	  (int) (((double) MINHIST *
		  pow ((double) (MAXHIST) / (double) MINHIST,
		       (double) (i - 1) / (double) ((NHIST - 1)))) + 0.5);
      }


//initialisation of index and tag computation functions
    for (int i = 1; i <= NHIST; i++)
      {
	ch_i[i].init (m[i], (logg[i]));
	ch_t[0][i].init (ch_i[i].OLENGTH, TB[i]);
	ch_t[1][i].init (ch_i[i].OLENGTH, TB[i] - 1);
      }
    for (int i = 1; i <= NHIST; i++)
      {
	ch_ios[i].init (m[i], (logg[i]));
	ch_tos[0][i].init (ch_i[i].OLENGTH, TB[i]);
	ch_tos[1][i].init (ch_i[i].OLENGTH, TB[i] - 1);
      }
    for (int i = 1; i <= NHIST; i++)
      {
	ch_i_pred[i].init (m[i], (logg[i]));
	ch_t_pred[0][i].init (ch_i[i].OLENGTH, TB[i]);
	ch_t_pred[1][i].init (ch_i[i].OLENGTH, TB[i] - 1);
      }
    for (int i = 1; i <= NHIST; i++)
      {
	ch_ios_pred[i].init (m[i], (logg[i]));
	ch_tos_pred[0][i].init (ch_i[i].OLENGTH, TB[i]);
	ch_tos_pred[1][i].init (ch_i[i].OLENGTH, TB[i] - 1);
      }

    for (int i = 1; i <= NHIST; i++)
      {
	for(int j = 0; j < (1 << (logg[i])); j++)
	  gtable[i][j].init();
      }

}

// the index functions for the tagged tables uses path history as in the OGEHL predictor
//F serves to mix path history
  int F (int A, int size, int bank)
  {
    int A1, A2;

    A = A & ((1 << size) - 1);
    A1 = (A & ((1 << logg[bank]) - 1));
    A2 = (A >> logg[bank]);
    A2 =
      ((A2 << bank) & ((1 << logg[bank]) - 1)) + (A2 >> (logg[bank] - bank));
    A = A1 ^ A2;
    A = ((A << bank) & ((1 << logg[bank]) - 1)) + (A >> (logg[bank] - bank));
    return (A);
  }
// gindex computes a full hash of pc, ghist and phist
  int gindex (address_t pc, int bank)
  {
    int index;
    int M = (m[bank] > 16) ? 16 : m[bank];
    index =
      pc ^ (pc >> (abs (logg[bank] - bank) + 1)) ^
      ch_i[bank].comp ^ F (phist, M, bank);

    //index = index ^ btable[bindex(pc)].lphr;

    return (index & ((1 << (logg[bank])) - 1));
  }

  //  tag computation
  uint16_t gtag (address_t pc, int bank)
  {
    int tag = pc ^ ch_t[0][bank].comp ^ (ch_t[1][bank].comp << 1);
    
    if((tag & ((1 << TB[bank]) - 1)) == 0)
      tag = pc ^ (ch_t[0][bank].comp)  ^ (ch_t[1][bank].comp << 2);

    return (tag & ((1 << TB[bank]) - 1));
  }
//index computation for kernel branchs 
  int gindexos (address_t pc, int bank)
  {
    int index;
    int M = (m[bank] > 16) ? 16 : m[bank];
    index =
      pc ^ (pc >> (abs (logg[bank] - bank) + 1)) ^ ch_ios[bank].
      comp ^ F (phistos, M, bank);

    
    return (index & ((1 << (logg[bank])) - 1));
  }

  //  tag computation for kernel branch
  uint16_t gtagos (address_t pc, int bank)
  {
    int tag = pc ^ ch_tos[0][bank].comp ^ (ch_tos[1][bank].comp << 1);
    
    if((tag & ((1 << TB[bank]) - 1)) == 0)
      tag = pc ^ (ch_tos[0][bank].comp )  ^ (ch_tos[1][bank].comp << 2);
    
    return (tag & ((1 << TB[bank]) - 1));

  }

// gindex computes a full hash of pc, ghist and phist
  int gindex_pred (address_t pc, int bank)
  {
    int index;
    int M = (m[bank] > 16) ? 16 : m[bank];
    index =
      pc ^ (pc >> (abs (logg[bank] - bank) + 1)) ^
      ch_i_pred[bank].comp ^ F (phist_pred, M, bank);

    return (index & ((1 << (logg[bank])) - 1));
  }

  //  tag computation
  uint16_t gtag_pred (address_t pc, int bank)
  {
    int tag = pc ^ ch_t_pred[0][bank].comp ^ (ch_t_pred[1][bank].comp << 1);
    
    if((tag & ((1 << TB[bank]) - 1)) == 0)
      tag = pc ^ (ch_t_pred[0][bank].comp)  ^ (ch_t_pred[1][bank].comp << 2);
     
    return (tag & ((1 << TB[bank]) - 1));
  }
//index computation for kernel branchs 
  int gindexos_pred (address_t pc, int bank)
  {
    int index;
    int M = (m[bank] > 16) ? 16 : m[bank];
    index =
      pc ^ (pc >> (abs (logg[bank] - bank) + 1)) ^ ch_ios_pred[bank].
      comp ^ F (phistos_pred, M, bank);

    return (index & ((1 << (logg[bank])) - 1));
  }

  //  tag computation for kernel branch
  uint16_t gtagos_pred (address_t pc, int bank)
  {

    int tag = pc ^ ch_tos_pred[0][bank].comp ^ (ch_tos_pred[1][bank].comp << 1);
    
    if((tag & ((1 << TB[bank]) - 1)) == 0)
      tag = pc ^ (ch_tos_pred[0][bank].comp)  ^ (ch_tos_pred[1][bank].comp << 2);

    return (tag & ((1 << TB[bank]) - 1));

  }

//just a simple pseudo random number generator: a 2-bit counter, used to avoid ping-pong phenomenom on tagged entry allocations

  int MYRANDOM ()
  {
    Seed++;
    return (Seed & 3);
  };

// shifting the global history:  we manage the history in a big table in order to reduce simulation time
  void updateghist (uint8_t * &h, bool dir, uint8_t * tab, int &PT)
  {
    if (PT == 0)
      {
	for (int i = 0; i < MAXHIST; i++)
	  tab[BUFFERHIST - MAXHIST + i] = tab[i];
	PT = BUFFERHIST - MAXHIST;
	h = &tab[PT];
      }
    PT--;
    h--;
    h[0] = (dir) ? 1 : 0;
  }

  // PREDICTION at the Fetch stage
  branch_update *predict_fetch (unsigned int pc)
  {

    bool KERNELMODE = ((pc & 0xc0000000) == 0xc0000000);

    //if (b->type & IS_BR_INDIRECT)
      {
	target_pred = 0;
// TAGE prediction

// computes the table addresses and the partial tags
        if (!KERNELMODE)
          for (int i = 1; i <= NHIST; i++)
            {
              GI[i] = gindex_pred (pc, i);
              GTAG[i] = gtag_pred (pc, i);
            }
        else
          for (int i = 1; i <= NHIST; i++)
            {
              GI[i] = gindexos_pred (pc, i);
              GTAG[i] = gtagos_pred (pc, i);
            }

        HitBank = 0;
        AltBank = 0;
	//Look for the bank with longest matching history
        for (int i = NHIST; i > 0; i--)
          {
            if (gtable[i][GI[i]].tag == GTAG[i])
              {
                HitBank = i;
                break;
              }
          }
	//check whether the branch hits in the cache_table
	int flag1=0;
	int rem_len=0;
	int rem_len1 = 0;
	for (int x=0;x<31;x++)
	{
	  if (pc==cache_table[x].pc)
	  {
	    flag1=1;
	    rem_len=cache_table[x].hist_len;
	    rem_len1 = rem_len;
	  }
	}

	//finding a table with shorter history lengths. If the branch hits cache_table, let its rem_len to determine which to use among multiple tables with shorter history lengths
      for (int i = HitBank - 1; i > 0; i--)
          {
	    if(flag1 == 0)
	    {
	      if (gtable[i][GI[i]].tag == GTAG[i] && gtable[i][GI[i]].alt == false)
		  {
		    AltBank = i;
		    break;
		  }
	    }
	    else
	    {
	      if (gtable[i][GI[i]].tag == GTAG[i] && gtable[i][GI[i]].alt == false)
	      {
		AltBank = i;
		if (rem_len!=1)
		{
		  rem_len--;
		}
		else
		{
		  break;
		}
	      }
	    }
	  }
      rem_len = rem_len1;
	//Look for the alternate bank
      if(AltBank == 0)
      {
        for (int i = HitBank - 1; i > 0; i--)
          {
            if (gtable[i][GI[i]].tag == GTAG[i])
              if ((USE_ALT_ON_NA < 0)) //USE_ALT_ON_NA is the legacy code from LTAGE. may not be useful at all.
                {
                  AltBank = i;
		  if (flag1==1)
		    {
		      if (rem_len!=1)
			{
			  rem_len--;
			}
		      else
			{
			  break;
			}
		    }
		  else 
		    {
		      break;
		    }
		  
                }
          }
      }
//computes the prediction and the alternate prediction
        if (HitBank > 0)
          {
            if (AltBank > 0)
		alttarget = gtable[AltBank][GI[AltBank]].targetaddr;
            else
	      alttarget = 0;
            if ((USE_ALT_ON_NA < 0))
		target_pred = gtable[HitBank][GI[HitBank]].targetaddr;
            else
		target_pred = alttarget;
          }
        else
          {
	    alttarget = 0;
	      target_pred = alttarget;
          }
	if(HitBank > 0)
	  {
	    if(gtable[HitBank][GI[HitBank]].alt && alttarget > 0)
	      target_pred = alttarget;   
	  }	 
      }
    u.target_prediction (target_pred);

    return &u;
  }
  void update_hist_fetch(const cbp3_uop_dynamic_t* b, bool taken) //update GHRs at the fetch stage, taken is the oracle information, whihc is used to update GHRs , not the prediction tables
  {
    bi = *b;
    int pc = b->pc;
 
//  UPDATE HISTORIES  

//check user and kernel mode to detect transition:

    bool KERNELMODE = ((pc & 0xc0000000) == 0xc0000000);

    bool TAKEN = ((!(bi.type & IS_BR_CONDITIONAL)) | (taken));  //HZ: Should not update GHR for unconditional and indirect branches
    
    bool PATHBIT = (bi.pc & 1);

    if (!KERNELMODE)
      {
//update user history
	updateghist (ghist_pred, TAKEN, GHIST_pred, ptghist_pred);
	phist_pred = (phist_pred << 1) + PATHBIT;

	phist_pred = (phist_pred & ((1 << 16) - 1));
//prepare next index and tag computations for user branchs 
	for (int i = 1; i <= NHIST; i++)
	  {
	    ch_i_pred[i].update (ghist_pred);
	    ch_t_pred[0][i].update (ghist_pred);
	    ch_t_pred[1][i].update (ghist_pred);
	  }
      }

// always update kernel history
    updateghist (ghistos_pred, TAKEN, GHISTOS_pred, ptghistos_pred);
    phistos_pred = (phistos_pred << 1) + PATHBIT;
    phistos_pred = (phistos_pred & ((1 << 16) - 1));
//prepare next index and tag computations for  kernel branchs
    for (int i = 1; i <= NHIST; i++)
      {
	ch_ios_pred[i].update (ghistos_pred);
	ch_tos_pred[0][i].update (ghistos_pred);
	ch_tos_pred[1][i].update (ghistos_pred);
      }
//END UPDATE HISTORIES

  }

 void update_hist_retire(const cbp3_uop_dynamic_t* b, bool taken) //update GHRs at the retire stage, taken is the oracle information, whihc is used to update GHRs , not the prediction tables
  {
    bi = *b;
    int pc = b->pc;
  
//  UPDATE HISTORIES  

//check user and kernel mode to detect transition:

    bool KERNELMODE = ((pc & 0xc0000000) == 0xc0000000);

    bool TAKEN = ((!(bi.type & IS_BR_CONDITIONAL)) | (taken));  //HZ: Should not update GHR for unconditional and indirect branches
 
    bool PATHBIT = (bi.pc & 1);

    if (!KERNELMODE)
      {
//update user history
	updateghist (ghist, TAKEN, GHIST, ptghist);
	phist = (phist << 1) + PATHBIT;

	phist = (phist & ((1 << 16) - 1));
//prepare next index and tag computations for user branchs 
	for (int i = 1; i <= NHIST; i++)
	  {
	    ch_i[i].update (ghist);
	    ch_t[0][i].update (ghist);
	    ch_t[1][i].update (ghist);
	  }
      }

// always update kernel history
    updateghist (ghistos, TAKEN, GHISTOS, ptghistos);
    phistos = (phistos << 1) + PATHBIT;
    phistos = (phistos & ((1 << 16) - 1));
//prepare next index and tag computations for  kernel branchs
    for (int i = 1; i <= NHIST; i++)
      {
	ch_ios[i].update (ghistos);
	ch_tos[0][i].update (ghistos);
	ch_tos[1][i].update (ghistos);
      }
//END UPDATE HISTORIES

  }

  // PREDICTION at the retire stage
  branch_update *predict_retire (const cbp3_uop_dynamic_t* b)
  {
    bi = *b;
    unsigned int pc = b->pc;

    bool KERNELMODE = ((pc & 0xc0000000) == 0xc0000000);

    if (b->type & IS_BR_INDIRECT)
      {
	target_pred = 0;
// TAGE prediction

// computes the table addresses and the partial tags
        if (!KERNELMODE)
          for (int i = 1; i <= NHIST; i++)
            {
              GI[i] = gindex (pc, i);
              GTAG[i] = gtag (pc, i);
            }
        else
          for (int i = 1; i <= NHIST; i++)
            {
              GI[i] = gindexos (pc, i);
              GTAG[i] = gtagos (pc, i);
            }
   
        HitBank = 0;
        AltBank = 0;
	//Look for the bank with longest matching history
        for (int i = NHIST; i > 0; i--)
          {
            if (gtable[i][GI[i]].tag == GTAG[i])
              {
                HitBank = i;
                break;
              }
          }

	//check whether the branch hits in the cache_table
	int flag1=0;
	int rem_len=0;
        for (int x=0;x<31;x++)
	  {
	    if (pc==cache_table[x].pc)
	      {
		flag1=1;
		
		if (cache_table[x].mspcount>2)
		  {	    
		    cache_table[x].mspcount=1;
		    
		  }
		rem_len=cache_table[x].hist_len;
		
	      }
	  }
	//finding a table with shorter history lengths. If the branch hits cache_table, let its rem_len to determine which to use among multiple tables with shorter history lengths
	//Look for the alternate bank
        for (int i = HitBank - 1; i > 0; i--)
          {
            if (gtable[i][GI[i]].tag == GTAG[i])
              if ((USE_ALT_ON_NA < 0))
                {
                  AltBank = i;
		  
		  if (flag1==1)
		    {
		      if (rem_len!=1)
			{   
			  break;
			}
		      else
			{
			  break;
			}
		    }
		  else
		    {
		      break;
		    }
                }
          }

	//computes the prediction and the alternate prediction
        if (HitBank > 0)
          {
            if (AltBank > 0)
      	      alttarget = gtable[AltBank][GI[AltBank]].targetaddr;
            else
	      alttarget = 0;//btable[BI].targetaddr;
	    //if the entry is recognized as a newly allocated entry and 
	    //USE_ALT_ON_NA is positive  use the alternate prediction
	    //Again, USE_ALT_ON_NA is from LTAGE. May not be used at all
            if ((USE_ALT_ON_NA < 0))
   	      target_pred = gtable[HitBank][GI[HitBank]].targetaddr;
            else
	      target_pred = alttarget;
          }
        else
          {
	    alttarget = 0;
	    target_pred = alttarget;
          }
        //end TAGE prediction
      }
    u.target_prediction (target_pred);
    return &u;
  }

  // PREDICTOR UPDATE
  void update_retire (branch_update * u, bool taken, unsigned int target, int n_br)
  {
    int NRAND = MYRANDOM ();
    address_t pc = bi.pc;
     ////////////////////////////////////////////////
    if (bi.type & IS_BR_INDIRECT)
      {

	// TAGE UPDATE  
	bool ALLOC = ((target_pred != target) & (HitBank < NHIST));
	
        if (HitBank > 0)
          {

	    unsigned int  LongestMatchTarget = (gtable[HitBank][GI[HitBank]].targetaddr);
            bool PseudoNewAlloc = 1;

	    if(HitBank > 0)
	    {
	      if(LongestMatchTarget == target)
		gtable[HitBank][GI[HitBank]].alt = false;
	      else if(alttarget == target)
		gtable[HitBank][GI[HitBank]].alt = true;
	      //setting whether to use long history length or not
	      if(LongestMatchTarget != target)
	      {
		int alti = 0;
		for (int i = HitBank - 1; i > 0; i--)
		{
		  if (gtable[i][GI[i]].tag == GTAG[i] && gtable[i][GI[i]].targetaddr == target)
		  {
		    alti = i;
		    break;
		  }
		}
		if(alti > 0)
		{
		  for(int i = HitBank; i > alti; i--)
		  {
		    if(gtable[i][GI[i]].tag == GTAG[i] && gtable[i][GI[i]].targetaddr != target)
		      gtable[i][GI[i]].alt = true;
		  }
		  gtable[alti][GI[alti]].alt = false;
		}
	      }	        	    
	    }	
	    
	    // an entry is considered as newly allocated if its prediction counter is weak
            if (PseudoNewAlloc)
              {
                if (LongestMatchTarget == target)
                  ALLOC = false;
		// if it was delivering the correct prediction, no need to allocate a new entry
		//even if the overall prediction was false

                if (LongestMatchTarget != alttarget)
                  {
                    if (alttarget == target)
                      {

                        if (USE_ALT_ON_NA < 7)
                          USE_ALT_ON_NA++;
                      }

                    else if (USE_ALT_ON_NA > -8)
                      USE_ALT_ON_NA--;

                  }
                if (USE_ALT_ON_NA >= 0)
                  target_pred = LongestMatchTarget;

              }

          }
	if (ALLOC)
	  {
	    // is there some "unuseful" entry to allocate
	    int8_t min = 1;
	    for (int i = NHIST; i > HitBank; i--)
	      if (gtable[i][GI[i]].u < min)
		min = gtable[i][GI[i]].u;
	    
	    // we allocate an entry with a longer history
	    //to  avoid ping-pong, we do not choose systematically the next entry, but among the 3 next entries
	    int Y = NRAND & ((1 << (NHIST - HitBank - 1)) - 1);
	    int X = HitBank + 1;
	    if (Y & 1)
	      {
		X++;
		if (Y & 2)
		  X++;
	      }
	    //NO ENTRY AVAILABLE:  ENFORCES ONE TO BE AVAILABLE 
	    if (min > 0)
	      gtable[X][GI[X]].u = 0;
     
	    
	    //Allocate only  one entry
	    
	    for (int i = X; i <= NHIST; i += 1)
	      
	      if ((gtable[i][GI[i]].u == 0))
		{
		  gtable[i][GI[i]].tag = GTAG[i];
		  gtable[i][GI[i]].targetaddr = target;
		  gtable[i][GI[i]].u = 0;
		  gtable[i][GI[i]].alt = false;
		  break;
		  
		}
	  }
	//periodic reset of u: reset is not complete but bit by bit
	TICK++;
	if ((TICK & ((1 << LOGTICK) - 1)) == 0)
	  // reset least significant bit
	  // most significant bit becomes least significant bit
	  for (int i = 1; i <= NHIST; i++)
	    for (int j = 0; j < (1 << logg[i]); j++)
	      gtable[i][j].u = gtable[i][j].u >> 1;
	  
	//use cache_table to keep hard-to-predict branches
	if (target_pred!= target)	
	  {
	    int flag2=0;
	    for (int loop1=0;loop1<32;loop1++)
	      {
		
		if (cache_table[loop1].pc==pc)//if an entry is present or not
		  {
		    flag2=1;
		    if (target_pred!=alttarget)
		      {
			cache_table[loop1].hist_len++;//change
			cache_table[loop1].mspcount++;
		      }
		  }
		
	      }
	    if (flag2==0 & (target_pred!=target))// if an entry is not present search for a table entry
	      {
		int lookfor=33;
		for (int loop2=0;loop2<32;loop2++)
		  {
		    if (cache_table[loop2].mspcount==0)
		      {
			lookfor=loop2;
			break;
		      }
		  }
		if (lookfor<33)
		  {
		    cache_table[lookfor].pc=pc;
		    cache_table[lookfor].mspcount=1;
		    cache_table[lookfor].hist_len=1;
		  }
	      }	
	  }
	// update the u counter
	if (target_pred != alttarget)
	  {
	    int flag2=0;
	    for (int loop1=0;loop1<32;loop1++)
	      {
		
		if (cache_table[loop1].pc==pc)//if an entry is present or not
		  {
		    flag2=1;
		    if (target_pred==target)
		      {
			if (cache_table[loop1].mspcount>2)
			  {
			    cache_table[loop1].mspcount=0;
			  }
			else
			  {
			    cache_table[loop1].mspcount=0;
			  }
		      }
		  }
		
	      }
	    if (flag2==0 & (target_pred!=target) & 0 )// if an entry is not present search for a table entry
	      {
		int lookfor=33;
		for (int loop2=0;loop2<32;loop2++)
		  {
		    if (cache_table[loop2].mspcount==0)
		      {
			lookfor=loop2;
			break;
		      }
		  }
		if (lookfor<33)
		  {
		    cache_table[lookfor].pc=pc;
		    cache_table[lookfor].mspcount=1;
		    cache_table[lookfor].hist_len=1;
		  }
	      }
	    if (target_pred == target)
	      {
		if (gtable[HitBank][GI[HitBank]].u < 3)
		  gtable[HitBank][GI[HitBank]].u++;
	      }
	    else
	      {
		if (USE_ALT_ON_NA < 0)
		  if (gtable[HitBank][GI[HitBank]].u > 0)
		    gtable[HitBank][GI[HitBank]].u--;
	      }
	    
	    
	  }
 
      }   
  }

  //using the address-target correlation table to make prediction when address (src4_value) is available
unsigned int pred_execute (const cbp3_uop_dynamic_t* branch, int src4_value)
{
  uint32_t epred=0;
  for (int i=0;i<EXEC_T_SIZE;i++)
  {
    if (branch->pc==exec_table[i].pc)
    {
      for (int j=0;j<NUM_PAIR;j++)
	{
	  if (src4_value==exec_table[i].sr[j])
	    {
	      epred=exec_table[i].pr[j];
	    }
	}
    }
  }
  return epred;
}

//update the the address-target correlation table when the branch target is known at the execution stage. The input parameter src4_value is the producer load address. 
 void pred_exe_update (const cbp3_uop_dynamic_t* branch, int src4_value,unsigned int target)
 {
   int hit=0;
   for (int j=0;j<EXEC_T_SIZE;j++)
     {
       if (branch->pc==exec_table[j].pc)
	 {
	   hit=1;

	   //update lru
	   for(int ii = 0; ii < EXEC_T_SIZE; ii++)
	     if(exec_table[ii].lru < exec_table[j].lru)
	       exec_table[ii].lru++;

	   exec_table[j].lru = 0;

	   int tick=0;
	   for (int k=0;k<NUM_PAIR;k++)
	     {
	       if (src4_value==exec_table[j].sr[k])
		 {
		   tick=1;
		   if (exec_table[j].pr[k]!=target)
		     {
		       //non-stable address-tragte correlation. reset the entry
		       exec_table[j].pc=0;
		       
		       for (int hcn=0;hcn<NUM_PAIR;hcn++)
			 {			   
			   exec_table[j].sr[hcn]=0;
			   exec_table[j].pr[hcn]=0;		   
			 }

		     }
		 }	
	     }
	   //address is not present in the address-target pairs
	   if (tick==0)
	     {
	       int tock = 0;
	       for (int k=0;k<NUM_PAIR;k++)
		 {
		   if (exec_table[j].pr[k]==0)
		     {
		       tock = 1;
		       exec_table[j].sr[k]=src4_value;
		       exec_table[j].pr[k]=target;
		       break;
		     }
		 }	
	       if(tock == 0)
	       {
		 //random replacement
		 int k = rand() % NUM_PAIR;//& 7;
		 exec_table[j].sr[k]=src4_value;
		 exec_table[j].pr[k]=target;
	       }
	     }
	   
	 }
       
     }	
   //miss in exec_table. Replace the LRU entry
 if (hit==0)
     {
       int y = -1;
       for(int ii = 0; ii < EXEC_T_SIZE; ii++)
	 if(exec_table[ii].lru == EXEC_T_SIZE-1)
	   y = ii;
       assert(y != -1);
       exec_table[y].pc=branch->pc;
       exec_table[y].sr[0]=src4_value;
       exec_table[y].pr[0]=target;
       for (int hcn=1;hcn<NUM_PAIR;hcn++)
	 {
	   exec_table[y].sr[hcn]=0;
	   exec_table[y].pr[hcn]=0;
	   
	 }
       for(int ii = 0; ii < EXEC_T_SIZE; ii++)
	 exec_table[ii].lru++;
       exec_table[y].lru = 0;
     }	
   
 }	
};
